#include <SoftwareSerial.h>
SoftwareSerial SSerial(2,3); // RX, TX

// No Extra bytes.
#define XBEE_TRANS_DATA       0x03  // Data Packet from Transmitter to Robot*
#define XBEE_ENTER_SSC_MODE     0x05  // The controller is letting robot know to enter SSC echo mode
#define XBEE_REQ_SN_NI        0x06  // Request the serial number and NI string


#define XBEE_RECV_DISP_VAL      0x84  // Display a value on line 2
// If <cbExtra> is  0 then we will display the number contained in <SerialNumber> 
// If not zero, then it is a count of bytes in a string to display.
#define XBEE_RECV_DISP_STR      0x85  // Display a string value on line 2
#define XBEE_PLAY_SOUND       0x86  // Will make sounds on the remote...
//  <cbExtra> - 2 bytes per sound: Duration <0-255>, Sound: <Freq/25> to make fit in byte...
#define XBEE_SSC_MODE_EXITED            0x87  // a message sent back to the controller when
// it has left SSC-mode.
// 4 bytes - SNH
// 4 bytes - SNL
// up to 20 bytes(probably 14) for NI


typedef struct _diypacketOrig {
  uint16_t  wButtons;         // the 16 buttons
  uint8_t bRJoyLR;          // right joystick X (LR)
  uint8_t bRJoyUD;          //        ..Y (UD)
  uint8_t bLJoyLR;          // Left Joystick  X (LR)
  uint8_t bLJoyUD;          //          Y (UD)
  uint8_t bRSlider;         // Right Slider
  uint8_t bLSlider;         // Left Slider
} 
DIYPACKETORIG;
// Main data packet            
typedef union {
  byte    ab[12];
  struct  {    
    uint16_t  wButtons;         // the 16 buttons
    byte  bRJoyLR;          // right joystick X (LR)
    byte  bRJoyUD;          //        ..Y (UD)
    byte  bLJoyLR;          // Left Joystick  X (LR)
    byte  bLJoyUD;          //          Y (UD)
    byte  bRSlider;         // Right Slider
    byte  bLSlider;         // Left Slider
    // Added values since first version                    
    byte    bRPot;                                          // top Pot on right joystick
    byte    bLPot;                                          // ... Left joystick
    byte  bMSlider;         // Middle Slider
    byte    bButtons2;                                      // Extra buttons on new DIY...
  } 
  s;
} 
DIYPACKET;

DIYPACKET  g_diyp;
DIYPACKET  g_diypPrev;
typedef DIYPACKET *PDIYPACKET;

typedef unsigned long ulong;

// Now define some static state stuff
typedef struct _diystate
{
  // state information
  boolean   fPacketValid; 
  boolean       fNewPacket;
  boolean       fDataPacketsReceived;
  boolean   fPacketTimeOut;
  uint8_t       cbPacketSize;    // what was the size of the last data packet received.
  // More internal to function...
  uint8_t bPacketNum;
  uint8_t       bTxStatusLast;                          // status of our last TX message.
  uint8_t bAPIPacket[33];       // Api packet
  uint16_t  wAPIDL;         // current destination.
  uint16_t      wDBGDL;                                // Debug Destination.

  // Other information, could make static to file...
  ulong         ulLastPacket;

} 
DIYSTATE;

DIYSTATE g_diystate;

unsigned long ulStart;


boolean ReceiveXBeePacket(PDIYPACKET pdiyp);

word   wCnt = 0;
word g_wCompositButtons;
boolean g_fFirstPacket = true; // On first packet should show all data.
word g_wAPIDLPrev;
word wDelay = 25;

uint8_t abT[10];

void setup() {
   SSerial.begin(57600);
   Serial.begin(115200);
   ClearXBeeInputBuffer();      

  delay(1000);

  // Lets try to get an OK 
  Serial.setTimeout(100);

  delay(2000);
  ClearXBeeInputBuffer();
  g_diystate.fPacketValid = false; 
  g_diystate.fDataPacketsReceived = false;
  g_diystate.bPacketNum = 0;
  g_diystate.wDBGDL= 0xffff;          // Debug 

  ulStart = millis();
}


//==============================================================================
// [ClearXBeeInputBuffer] - This simple helper function will clear out the input
//                        buffer from the XBEE
//==============================================================================
void ClearXBeeInputBuffer(void)
{
  uint8_t b[1];

 
      Serial.flush();    // clear out anything that was queued up...        
  while (Serial.read() != -1)
    ;    // just loop as long as we receive something...

}

char aa='0';
boolean bInit=false;
char g_szButton[] = "  Pressed";
void loop() {
  
  
  if (ReceiveXBeePacket(&g_diyp)) {
    
     uint16_t wCompositButtons = g_diyp.s.wButtons;
    if(g_diystate.fNewPacket) {
      while (g_diystate.fNewPacket) {
        wCnt++;  // Count of messages
        wCompositButtons = (uint16_t)(wCompositButtons & (uint16_t)~g_diypPrev.s.wButtons) | (uint16_t)g_diyp.s.wButtons;
        if (!ReceiveXBeePacket(&g_diyp))
          break;
      }
      

      if (memcmp((void*)&g_diyp, (void*)&g_diypPrev, sizeof(g_diyp)) != 0) {
        if (g_fFirstPacket || g_diyp.s.wButtons != g_diypPrev.s.wButtons) {
          //DisplayField(FIELD_X, FIELD_Y1 + 0 * FIELD_YINC, g_diyp.s.wButtons, HEX);
          if (g_diyp.s.wButtons && !g_diypPrev.s.wButtons) {
            // not 100% good, but fine for test. 
            uint16_t w = g_diyp.s.wButtons;
            char ch = '0';
            while ((w & 1) == 0) {
              w>>= 1;
              ch = (ch == '9')? 'A' : (ch + 1);
            }
            g_szButton[0] = ch;
            SSerial.print(g_szButton);SSerial.println();
          }   
        }  
        
        if (g_fFirstPacket || (g_diyp.s.bRJoyLR != g_diypPrev.s.bRJoyLR))
          {SSerial.print("RJoyLR:");SSerial.print( g_diyp.s.bRJoyLR);SSerial.println();}
        if ((g_fFirstPacket || g_diyp.s.bRJoyUD != g_diypPrev.s.bRJoyUD))
          {SSerial.print("RJoyUD:");SSerial.print(g_diyp.s.bRJoyUD);SSerial.println();}
        if (g_fFirstPacket || (g_diyp.s.bLJoyLR != g_diypPrev.s.bLJoyLR))
          {SSerial.print("RLJoyLR");SSerial.print( g_diyp.s.bLJoyLR);SSerial.println();}
        if (g_fFirstPacket || (g_diyp.s.bLJoyUD != g_diypPrev.s.bLJoyUD))
          {SSerial.print("LJoyUD");SSerial.print( g_diyp.s.bLJoyUD);SSerial.println();}
        /*if (g_fFirstPacket || (g_diyp.s.bRSlider != g_diypPrev.s.bRSlider))
          {DisplayField(FIELD_X, FIELD_Y1 + 5 * FIELD_YINC, g_diyp.s.bRSlider);}
        if (g_fFirstPacket || (g_diyp.s.bLSlider != g_diypPrev.s.bLSlider))
          {DisplayField(FIELD_X, FIELD_Y1 + 6 * FIELD_YINC, g_diyp.s.bLSlider);}
        */if (g_diystate.cbPacketSize > sizeof(DIYPACKETORIG)) {
          if (g_fFirstPacket || (g_diyp.s.bRPot != g_diypPrev.s.bRPot)) 

            // Note Not all remotes have these fields.
          //  DisplayField(FIELD_X, FIELD_Y1 + 7 * FIELD_YINC, g_diyp.s.bRPot);
          //if (g_fFirstPacket || (g_diyp.s.bLPot != g_diypPrev.s.bLPot))
          //  DisplayField(FIELD_X, FIELD_Y1 + 8 * FIELD_YINC, g_diyp.s.bLPot);


          if (g_diystate.cbPacketSize > (sizeof(DIYPACKETORIG)+2)) {
           // if (g_fFirstPacket || (g_diyp.s.bMSlider != g_diypPrev.s.bMSlider)) 
           //   DisplayField(FIELD_X, FIELD_Y1 + 9 * FIELD_YINC, g_diyp.s.bMSlider);
          }
        }
        if (g_diystate.wAPIDL != g_wAPIDLPrev){
          g_wAPIDLPrev = g_diystate.wAPIDL;
      //    DisplayField(FIELD_X, FIELD_Y1 + 11 * FIELD_YINC, g_wAPIDLPrev, HEX);
        }
        // remember current state
        g_diypPrev = g_diyp;
        g_fFirstPacket = false;
      }
    }
  }      

  unsigned long ulTime = millis();
  if ((ulTime - ulStart) >= 1000) {
    //DisplayField(FIELD_X, FIELD_Y1 + 12 * FIELD_YINC, wCnt/*(wCnt*1000)/(ulTime - ulStart)*/);
//    XBeeOutputVal((wCnt*1000)/(ulTime-ulStart));
    ulStart = ulTime;
    if (!wCnt && !g_fFirstPacket) {
      // We displayed stuff but nothing coming in, lets clear display...
      g_fFirstPacket = true;
    }
    wCnt = 0;
  }
       
  

  delay(wDelay);  // put some delay in to see about reading multiple packets.

  
 
}



//==============================================================================
// [ReceiveXBeePacket] - This function will try to receive a packet of information
//         from the remote control over XBee.
//
// the data in a standard packet is arranged in the following byte order:
//    0 - Buttons High
//    1 - Buttons Low
//     2 - Right Joystick L/R
//    3 - Right Joystick U/D
//    4 - Left Joystick L/R
//    5 - Left Joystick U/D
//     6 - Right Slider
//    7 - Left Slider
//==============================================================================
boolean ReceiveXBeePacket(PDIYPACKET pdiyp)
{
  uint8_t cbRead;
  uint8_t bDataOffset;
  uint16_t wNewDL;
  ulong ulCurrentTime;
  ulong ulTimeDiffMS;
  boolean _fPacketValidPrev = g_diystate.fPacketValid;        // Save away the previous state as this is the state if no new data...
  boolean _fNewPacketAvail = false;

  g_diystate.fPacketValid = false;
  g_diystate.fNewPacket = false;      // Did we receive a new packet?

  //    We will first see if we have a packet header waiting for us.
  //  BUGBUG:: Question should I loop after I process a package and only get out of the
  //             loop when I have no more data, or only process one possible message?
  //            Maybe depends on message?
  for (;;)
  {
    if (!Serial.available())
      break;        // no input available, break from this loop

    // The XBEE has sent us some data so try to get a packet header
    cbRead = APIRecvPacket(10);        // Lets read in a complete packet.
    
    if (!cbRead)
      break;                            // Again nothing read?                                                                      ICI

    if (g_diystate.bAPIPacket[0] ==0x81)
      {bDataOffset = 5;  }             // Received packet with 16 bit addressing
    else if (g_diystate.bAPIPacket[0] == 0x80)
      bDataOffset = 11;                // Received packet with 64 bit addressing
    else if (g_diystate.bAPIPacket[0] == 0x89) {
      if (g_diystate.bAPIPacket[2] != g_diystate.bTxStatusLast) {
        g_diystate.bTxStatusLast = g_diystate.bAPIPacket[2];

      }
      continue;
    }
    else
      break;                            // Invalid packet lets bail from this loop.

    // Change CB into the number of extra bytes...
    cbRead -= (bDataOffset + 1);        // Ph is only 1 byte long now... 

    //-----------------------------------------------------------------------------
    // [XBEE_TRANS_DATA]
    //-----------------------------------------------------------------------------
    if (g_diystate.bAPIPacket[bDataOffset + 0] == XBEE_TRANS_DATA) {
      if (cbRead >= sizeof(DIYPACKETORIG)) {
        pdiyp->s.wButtons = g_diystate.bAPIPacket[bDataOffset + 1] + (g_diystate.bAPIPacket[bDataOffset + 2] << 8);

        // Simple copy the memory down
        g_diystate.cbPacketSize = (cbRead <= sizeof(DIYPACKET))? cbRead : sizeof(DIYPACKET);

        memcpy(&pdiyp->ab[2], &g_diystate.bAPIPacket[bDataOffset + 3], g_diystate.cbPacketSize);

        // process first as higher number of these come in...
        g_diystate.fPacketValid = true;    // data is valid
        g_diystate.fNewPacket = true;      // Did we receive a new packet?
        g_diystate.fDataPacketsReceived = true;
        // BUGBUG:: Assuming 16 bit addressing
        g_diystate.wAPIDL = (uint16_t)(g_diystate.bAPIPacket[1] << 8) + g_diystate.bAPIPacket[2];
        g_diystate.ulLastPacket = millis();
        return true;    //          // get out quick!
      } 
      else {

        SSerial.print("XBEE_TRANS_DATA: ");
        SSerial.print(sizeof(DIYPACKET), DEC);
        SSerial.print(" ");
        SSerial.println(cbRead, DEC);

      }
    }

    //-----------------------------------------------------------------------------
    // [UNKNOWN PACKET]
    //-----------------------------------------------------------------------------
    else {

      
        SSerial.print("Unknown Packet: ");
        SSerial.print(g_diystate.bAPIPacket[bDataOffset + 0], HEX);
        SSerial.print(" ");
        SSerial.println(cbRead, DEC);
      

    }
  }

  //-----------------------------------------------------------------------------
  // Exited above loop now See if we we are in some form of timeout condition.
  // We should receive a new data message about 30 times per second.  If we
  // have gone for a reasonably long time without receiving a message, assume
  // transmitter went away!
  //-----------------------------------------------------------------------------
  // Only send when we know the transmitter is ready.  Also if we are in the New data only mode don't ask for data unless we have been 
  //    old there
  // is new data. We relax this a little and be sure to ask for data every so often as to make sure the remote is still working...
  // 
  if (g_diystate.fDataPacketsReceived) {
    ulCurrentTime = millis();
    ulTimeDiffMS = ulCurrentTime - g_diystate.ulLastPacket;

    // See if we exceeded a global timeout.  If so let caller know so they can stop themself if necessary...
    if (ulTimeDiffMS > 250) {
      g_diystate.fPacketValid = 0;
      return false;
    }

    g_diystate.fPacketValid = _fPacketValidPrev;    // Say the data is in the same state as the previous call...
  }

  return g_diystate.fPacketValid;
}





//////////////////////////////////////////////////////////////////////////////
//==============================================================================
// [APIRecvPacket - try to receive a packet from the XBee. 
//        - Will return The packet length if it receives something, else 0
//        - Pass in buffer to receive packet.  Assumed it is big enough...
//        - pass in timeout if zero will return if no data...
//        
//==============================================================================
byte APIRecvPacket(ulong Timeout)
{
  byte cbRead;
  byte abT[3];
  byte bChksum;
  int i;

  short wPacketLen;
  if (Timeout == 0) 
  {
    if (!Serial.available())
      return 0;        // nothing waiting for us...
    Timeout = 100;    // .1 second?

  }

  Serial.setTimeout(Timeout);
  

  // Now lets try to read in the data from the xbee...
  // first read in the delimter and packet length
  // We now do this in two steps.  The first to try to resync if the first character
  // is not the proper delimiter...

  do {    
    cbRead = Serial.readBytes(abT, 1);
    if (cbRead == 0)
      {return 0;      }
      //else{SSerial.print(abT[0]);}
  } 
  while (abT[0] != 0x7e);//0x7e


 
  cbRead = Serial.readBytes(abT, 2);
  if (cbRead != 2)
    return 0;                // did not read in full header or the header was not correct.

  wPacketLen = (abT[0] << 8) + abT[1];

  if (wPacketLen >= sizeof(g_diystate.bAPIPacket)) {
     
    ClearXBeeInputBuffer();  // maybe clear everything else out.
    return 0;  // Packet bigger than expected maybe bytes lost...
  }
  
  // Now lets try to read in the packet
  cbRead = Serial.readBytes(g_diystate.bAPIPacket, wPacketLen+1);


  // Now lets verify the checksum.
  bChksum = 0;
  for (i = 0; i < wPacketLen; i++)
    bChksum = bChksum + g_diystate.bAPIPacket[i];             // Add that byte to the buffer...


  if (g_diystate.bAPIPacket[wPacketLen] != (0xff - bChksum)) {
       
    ClearXBeeInputBuffer();  // maybe clear everything else out.
    return 0;                // checksum was off
  }
  
  return wPacketLen;    // return the packet length as the caller may need to know this...
}
